import typing as T

import numpy as np
from scipy.special import ive

try:
    import torch
except ImportError:
    torch = None


def compute_chebychev_coeff_all_torch(eigval, t, K):
    with torch.no_grad():
        eigval = eigval.detach().cpu()
        out = 2.0 * ive(torch.arange(0, K + 1), -t * eigval)
    return out


def compute_chebychev_coeff_all(eigval, t, K):
    return 2.0 * ive(
        np.arange(
            0,
            K + 1,
        ),
        -t * eigval,
    )


def expm_multiply(
    L: np.ndarray,
    X: np.ndarray,
    coeff: np.ndarray,
    eigval: np.ndarray,
):
    """Matrix exponential with Chebyshev polynomial approximation."""

    def body(carry, c):
        T0, T1, Y = carry
        T2 = (2.0 / eigval) * (L @ T1) - 2.0 * T1 - T0
        Y = Y + c * T2
        return (T1, T2, Y)

    T0 = X
    Y = 0.5 * coeff[0] * T0
    T1 = (1.0 / eigval) * (L @ X) - T0
    Y = Y + coeff[1] * T1

    initial_state = (T0, T1, Y)
    for c in coeff[2:]:
        initial_state = body(initial_state, c)

    _, _, Y = initial_state

    return Y



import pytest
import numpy as np
from geosink.heat_kernel import HeatFilter, laplacian_from_data
import scipy
from scipy.sparse import rand

def gt_heat_kernel_data(
    data,
    t,
    sigma,
    alpha=20,
):
    L = laplacian_from_data(data, sigma, alpha=alpha)
    # eigendecomposition
    eigvals, eigvecs = np.linalg.eigh(L)
    # compute the heat kernel
    heat_kernel = eigvecs @ np.diag(np.exp(-t * eigvals)) @ eigvecs.T
    heat_kernel = (heat_kernel + heat_kernel.T) / 2
    heat_kernel[heat_kernel < 0] = 0.0
    return heat_kernel


def test_laplacian():
    data = np.random.normal(0, 1, (100, 5))
    sigma = 1.0
    L = laplacian_from_data(data, sigma)
    assert np.allclose(L, L.T)
    # compute the largest eigenvalue
    eigvals, _ = np.linalg.eigh(L)
    max_eigval = eigvals.max()
    assert max_eigval <= 2.0


@pytest.mark.parametrize("t", [0.1, 1.0, 10.0])
@pytest.mark.parametrize("order", [10, 30, 50])
@pytest.mark.parametrize("method", ["cheb", "exact"])
def test_heat_kernel_gaussian(t, order, method):
    data = np.random.normal(0, 1, (100, 5))
    lap = laplacian_from_data(data, sigma=1.0)
    heat_op = HeatFilter(lap=lap, tau=t, order=order, method=method)
    heat_kernel = heat_op.heat_kernel

    # test if symmetric
    assert np.allclose(heat_kernel, heat_kernel.T)

    # test if positive
    assert np.all(heat_kernel >= 0)

    # test if the heat kernel is close to the ground truth
    gt_heat_kernel = gt_heat_kernel_data(data, t=t, sigma=1.0)
    assert np.allclose(heat_kernel, gt_heat_kernel, atol=1e-1, rtol=1e-1)

@pytest.mark.limit_memory("800 MB")
def test_sparse_heat_kernel():
    t = 5.0
    order = 10
    adj = rand(10_000, 10_000, density=0.0001, format="coo", dtype=np.float32)
    lap = scipy.sparse.csgraph.laplacian(adj, symmetrized=False)

    heat_op = HeatFilter(lap=lap, tau=t, order=order, method="cheb")
    signal = np.random.uniform(0, 1, (10_000,1))
    signal = signal / signal.sum()
    diffused_signal = heat_op(signal)

    np.testing.assert_array_equal(np.isfinite(diffused_signal), True)

if __name__ == "__main__":
    pytest.main([__file__])


import numpy as np

from geosink.heat_kernel import HeatFilter


class GeoSinkhorn:
    def __init__(self, tau, order, method, graph=None, lap=None):
        assert (graph is not None) or (
            lap is not None
        ), "Either graph or lap must be provided."
        self.tau = tau
        kwargs_filter = {"lap": lap} if graph is None else {"graph": graph}
        self.heat_filter = HeatFilter(
            tau=tau, order=order, method=method, **kwargs_filter
        )

    def __call__(
        self,
        m_0,
        m_1,
        stopThr=1e-4,
        max_iter=1e3,
        verbose=False,
        return_vw=False,
        stopping_crit=False,
        stopping_fn=lambda x: False,
    ):
        eps = 1e-8
        assert m_0.shape == m_1.shape, "m_0 and m_1 must have the same shape"
        N = m_0.shape[0]
        v = np.ones(N)
        w = np.ones(N)
        a = np.ones(N) / N
        for i in range(1, int(max_iter) + 1):
            v_prev = v
            w_prev = w
            v = m_0 / (self.heat_filter(a * w) + eps)
            w = m_1 / (self.heat_filter(a * v) + eps)
            if (
                np.any(np.isnan(v))
                or np.any(np.isnan(w))
                or np.any(np.isinf(v))
                or np.any(np.isinf(w))
            ):
                v = v_prev
                w = w_prev
                break
            if i % 100 == 0:
                if verbose:
                    print(i, np.sum(np.abs(v - v_prev)))
                if np.sum(np.abs(v - v_prev)) < stopThr:
                    if verbose:
                        print("converged at iteration %d" % i)
                    break
            if stopping_crit:
                kl = np.sum(
                    4
                    * self.tau
                    * a
                    * (m_0 * np.log(v + eps) + m_1 * np.log(w + eps))
                )
                if stopping_fn(kl):
                    print("converged at iteration %d" % i)
                    break
        kl = np.sum(
            4 * self.tau * a * (m_0 * np.log(v + eps) + m_1 * np.log(w + eps))
        )
        if return_vw:
            return kl, v, w
        return kl



from typing import Any, Callable, Optional

import numpy as np
import scipy
import scipy.sparse.linalg
from scipy.sparse.linalg import eigsh

from geosink.cheb_approx import compute_chebychev_coeff_all, expm_multiply

try:
    import pygsp
except ImportError:
    pygsp = None

EPS_LOG = 1e-6
EPS_HEAT = 1e-4


def norm_sym_laplacian(A: np.array):
    deg = A.sum(axis=1)
    deg_sqrt_inv = np.diag(1.0 / np.sqrt(deg + EPS_LOG))
    id = np.eye(A.shape[0])
    return id - deg_sqrt_inv @ A @ deg_sqrt_inv


def laplacian_from_data(data: np.array, sigma: float, alpha: int = 20):
    affinity = np.exp(
        -((scipy.spatial.distance.cdist(data, data) / (2 * sigma)) ** alpha)
    )
    return norm_sym_laplacian(affinity)


class HeatFilter:
    """Wrapper for the approximation of the heat kernel."""

    _valid_methods = ["cheb_pygsp", "cheb", "lowrank", "exact"]

    def __init__(
        self,
        tau: float,  # Diffusion time
        order: int,  # Degree or numver of steps
        method: str,  # filter `" cheb_pygsp"`, `"cheb"`, `"lowrank"`, `"exact"`.
        graph: Optional[Any] = None,  # Graph object
        lap: Optional[np.array] = None,  # Laplacian matrix
    ) -> Callable:
        self.graph = graph
        self.tau = tau
        self.order = order
        self.method = method

        assert (graph is not None) or (
            lap is not None
        ), "Either graph or lap must be provided."
        self.lap = graph.L if lap is None else lap

        if method not in self._valid_methods:
            raise ValueError(
                "method must be one of {}".format(self._valid_methods)
            )

        if method == "cheb_pygsp":
            assert (
                graph is not None
            ), "graph must be provided for method cheb_pygsp"
            graph.estimate_lmax()
            self._filter = pygsp.filters.Heat(graph, tau)

        elif method == "cheb":
            self.phi = eigsh(self.lap, k=1, return_eigenvectors=False)[0] / 2
            self.coeff = compute_chebychev_coeff_all(
                self.phi, self.tau, self.order
            )

        elif method == "lowrank":
            self.lap.shape[0]
            eval, evec = eigsh(self.lap, k=order, which="SM")
            self._filter = evec @ np.diag(np.exp(-self.tau * eval)) @ evec.T

        elif method == "exact":
            eval, evec = scipy.linalg.eigh(self.lap)
            self._filter = evec @ np.diag(np.exp(-self.tau * eval)) @ evec.T

    @property
    def heat_kernel(self):
        heat_kernel = self(np.eye(self.lap.shape[0]))
        heat_kernel = (heat_kernel + heat_kernel.T) / 2
        heat_kernel[heat_kernel < 0] = 0.0
        return heat_kernel

    def __call__(self, b, safe_zeros=True):

        if self.method == "cheb_pygsp":
            diff = self._filter.filter(b, order=self.order)

        elif self.method == "cheb":
            diff = expm_multiply(self.lap, b, self.coeff, self.phi)

        elif self.method in ["lowrank", "exact"]:
            diff = self._filter @ b

        if safe_zeros:
            diff[diff < 0.0] = 0.0
        return diff



from geosink.sinkhorn import GeoSinkhorn 
from geosink.heat_kernel import laplacian_from_data
import numpy as np
import pytest
import scipy
from scipy.sparse import rand


@pytest.mark.parametrize("order", [10, 30, 50])
@pytest.mark.parametrize("num_iter", [100, 200, 300])
@pytest.mark.parametrize("tau", [0.1, 1.0, 10.0])
def test_geo_sinkhorn(order, num_iter, tau):
    data = np.random.normal(0, 1, (1000, 5))
    lap = laplacian_from_data(data, sigma=1.0)
    conv_sinkhorn = GeoSinkhorn(tau=tau, order=order, method="cheb", lap=lap)
    m_0 = np.random.rand(1000,)
    m_0[:500] = 0
    m_0 = m_0 / np.sum(m_0)
    m_1 = np.random.rand(1000,)
    m_1[500:] = 0
    m_1 = m_1 / np.sum(m_1)
    kl, v, w = conv_sinkhorn(m_0, m_1, return_vw=True, max_iter=num_iter) 
    assert np.isfinite(kl)
    assert kl >= 0
    assert np.all(np.isfinite(v))
    assert np.all(np.isfinite(w))
 
def test_sparse_geo_sinkhorn():
    adj = rand(10_000, 10_000, density=0.0001, format="coo", dtype=np.float32)
    lap = scipy.sparse.csgraph.laplacian(adj, symmetrized=False)
    conv_sinkhorn = GeoSinkhorn(tau=1.0, order=10, method="cheb", lap=lap)
    m_0 = np.random.rand(10_000,)
    m_0[:5_000] = 0
    m_0 = m_0 / np.sum(m_0)
    m_1 = np.random.rand(10_000,)
    m_1[5_000:] = 0
    m_1 = m_1 / np.sum(m_1)
    kl, v, w = conv_sinkhorn(m_0, m_1, return_vw=True, max_iter=100)
    assert np.isfinite(kl)
    assert kl >= 0
    assert np.all(np.isfinite(v))
    assert np.all(np.isfinite(w))

def test_gaussian():
    # random normal data
    data0 = np.random.normal(0, 1, (100, 5))
    data1 = np.random.normal(5, 1, (100, 5))
    data2 = np.random.normal(10, 1, (100, 5))
    data = np.concatenate([data0, data1, data2], axis=0)
    lap = laplacian_from_data(data, sigma=1.0)
    conv_sinkhorn = GeoSinkhorn(tau=1.0, order=10, method="cheb", lap=lap)
    m_0 = np.zeros(300,)
    m_0[:100] = 1
    m_0 = m_0 / np.sum(m_0)
    m_1 = np.zeros(300,)
    m_1[100:200] = 1
    m_1 = m_1 / np.sum(m_1)
    m_2 = np.zeros(300,)
    m_2[200:] = 1
    kl = conv_sinkhorn(m_0, m_1, max_iter=500)
    kl2 = conv_sinkhorn(m_0, m_2, max_iter=500)
    assert kl2 > kl

if __name__ == "__main__":
    pytest.main([__file__])